@@ -1,4 +1,4 @@
-# Copyright 2009-2012 10gen, Inc.
+# Copyright 2009-2014 MongoDB, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -14,6 +14,8 @@
 
 """Database level operations."""
 
+import warnings
+
 from bson.binary import OLD_UUID_SUBTYPE
 from bson.code import Code
 from bson.dbref import DBRef
@@ -21,10 +23,12 @@ from bson.son import SON
 from pymongo import auth, common, helpers
 from pymongo.collection import Collection
 from pymongo.errors import (CollectionInvalid,
+                            ConfigurationError,
                             InvalidName,
                             OperationFailure)
-from pymongo.son_manipulator import ObjectIdInjector
-from pymongo import read_preferences as rp
+from pymongo.read_preferences import (modes,
+                                      secondary_ok_commands,
+                                      ReadPreference)
 
 
 def _check_name(name):
@@ -64,13 +68,15 @@ class Database(common.BaseObject):
                              secondary_acceptable_latency_ms=(
                                  connection.secondary_acceptable_latency_ms),
                              safe=connection.safe,
+                             uuidrepresentation=connection.uuid_subtype,
                              **connection.write_concern)
 
         if not isinstance(name, basestring):
             raise TypeError("name must be an instance "
                             "of %s" % (basestring.__name__,))
 
-        _check_name(name)
+        if name != '$external':
+            _check_name(name)
 
         self.__name = unicode(name)
         self.__connection = connection
@@ -79,7 +85,6 @@ class Database(common.BaseObject):
         self.__incoming_copying_manipulators = []
         self.__outgoing_manipulators = []
         self.__outgoing_copying_manipulators = []
-        self.add_son_manipulator(ObjectIdInjector())
 
     def add_son_manipulator(self, manipulator):
         """Add a new son manipulator to this database.
@@ -242,6 +247,16 @@ class Database(common.BaseObject):
 
         return Collection(self, name, **opts)
 
+    def _apply_incoming_manipulators(self, son, collection):
+        for manipulator in self.__incoming_manipulators:
+            son = manipulator.transform_incoming(son, collection)
+        return son
+
+    def _apply_incoming_copying_manipulators(self, son, collection):
+        for manipulator in self.__incoming_copying_manipulators:
+            son = manipulator.transform_incoming(son, collection)
+        return son
+
     def _fix_incoming(self, son, collection):
         """Apply manipulators to an incoming SON object before it gets stored.
 
@@ -249,10 +264,8 @@ class Database(common.BaseObject):
           - `son`: the son object going into the database
           - `collection`: the collection the son object is being saved in
         """
-        for manipulator in self.__incoming_manipulators:
-            son = manipulator.transform_incoming(son, collection)
-        for manipulator in self.__incoming_copying_manipulators:
-            son = manipulator.transform_incoming(son, collection)
+        son = self._apply_incoming_manipulators(son, collection)
+        son = self._apply_incoming_copying_manipulators(son, collection)
         return son
 
     def _fix_outgoing(self, son, collection):
@@ -268,9 +281,80 @@ class Database(common.BaseObject):
             son = manipulator.transform_outgoing(son, collection)
         return son
 
+    def _command(self, command, value=1,
+                 check=True, allowable_errors=None,
+                 uuid_subtype=OLD_UUID_SUBTYPE, compile_re=True, **kwargs):
+        """Internal command helper.
+        """
+
+        if isinstance(command, basestring):
+            command = SON([(command, value)])
+
+        command_name = command.keys()[0].lower()
+        must_use_master = kwargs.pop('_use_master', False)
+        if command_name not in secondary_ok_commands:
+            must_use_master = True
+
+        # Special-case: mapreduce can go to secondaries only if inline
+        if command_name == 'mapreduce':
+            out = command.get('out') or kwargs.get('out')
+            if not isinstance(out, dict) or not out.get('inline'):
+                must_use_master = True
+
+        # Special-case: aggregate with $out cannot go to secondaries.
+        if command_name == 'aggregate':
+            for stage in kwargs.get('pipeline', []):
+                if '$out' in stage:
+                    must_use_master = True
+                    break
+
+        extra_opts = {
+            'as_class': kwargs.pop('as_class', None),
+            'slave_okay': kwargs.pop('slave_okay', self.slave_okay),
+            '_must_use_master': must_use_master,
+            '_uuid_subtype': uuid_subtype
+        }
+
+        extra_opts['read_preference'] = kwargs.pop(
+            'read_preference',
+            self.read_preference)
+        extra_opts['tag_sets'] = kwargs.pop(
+            'tag_sets',
+            self.tag_sets)
+        extra_opts['secondary_acceptable_latency_ms'] = kwargs.pop(
+            'secondary_acceptable_latency_ms',
+            self.secondary_acceptable_latency_ms)
+        extra_opts['compile_re'] = compile_re
+
+        fields = kwargs.get('fields')
+        if fields is not None and not isinstance(fields, dict):
+            kwargs['fields'] = helpers._fields_list_to_dict(fields)
+
+        command.update(kwargs)
+
+        # Warn if must_use_master will override read_preference.
+        if (extra_opts['read_preference'] != ReadPreference.PRIMARY and
+                extra_opts['_must_use_master']):
+            warnings.warn("%s does not support %s read preference "
+                          "and will be routed to the primary instead." %
+                          (command_name,
+                           modes[extra_opts['read_preference']]),
+                          UserWarning, stacklevel=3)
+
+        cursor = self["$cmd"].find(command, **extra_opts).limit(-1)
+        for doc in cursor:
+            result = doc
+
+        if check:
+            msg = "command %s failed: %%s" % repr(command).replace("%", "%%")
+            helpers._check_command_response(result, self.connection.disconnect,
+                                            msg, allowable_errors)
+
+        return result, cursor.conn_id
+
     def command(self, command, value=1,
                 check=True, allowable_errors=[],
-                uuid_subtype=OLD_UUID_SUBTYPE, **kwargs):
+                uuid_subtype=OLD_UUID_SUBTYPE, compile_re=True, **kwargs):
         """Issue a MongoDB command.
 
         Send command `command` to the database and return the
@@ -315,6 +399,12 @@ class Database(common.BaseObject):
             in this list will be ignored by error-checking
           - `uuid_subtype` (optional): The BSON binary subtype to use
             for a UUID used in this command.
+          - `compile_re` (optional): if ``False``, don't attempt to compile
+            BSON regular expressions into Python regular expressions. Return
+            instances of :class:`~bson.regex.Regex` instead. Can avoid
+            :exc:`~bson.errors.InvalidBSON` errors when receiving
+            Python-incompatible regular expressions, for example from
+            ``currentOp``
           - `read_preference`: The read preference for this connection.
             See :class:`~pymongo.read_preferences.ReadPreference` for available
             options.
@@ -332,6 +422,10 @@ class Database(common.BaseObject):
           - `**kwargs` (optional): additional keyword arguments will
             be added to the command document before it is sent
 
+        .. note:: ``command`` ignores the ``network_timeout`` parameter.
+
+        .. versionchanged:: 2.7
+           Added ``compile_re`` option.
         .. versionchanged:: 2.3
            Added `tag_sets` and `secondary_acceptable_latency_ms` options.
         .. versionchanged:: 2.2
@@ -347,61 +441,22 @@ class Database(common.BaseObject):
         .. mongodoc:: commands
         .. _localThreshold: http://docs.mongodb.org/manual/reference/mongos/#cmdoption-mongos--localThreshold
         """
+        return self._command(command, value, check, allowable_errors,
+                             uuid_subtype, compile_re, **kwargs)[0]
 
-        if isinstance(command, basestring):
-            command = SON([(command, value)])
-
-        command_name = command.keys()[0].lower()
-        must_use_master = kwargs.pop('_use_master', False)
-        if command_name not in rp.secondary_ok_commands:
-            must_use_master = True
-
-        # Special-case: mapreduce can go to secondaries only if inline
-        if command_name == 'mapreduce':
-            out = command.get('out') or kwargs.get('out')
-            if not isinstance(out, dict) or not out.get('inline'):
-                must_use_master = True
-
-        extra_opts = {
-            'as_class': kwargs.pop('as_class', None),
-            'slave_okay': kwargs.pop('slave_okay', self.slave_okay),
-            '_must_use_master': must_use_master,
-            '_uuid_subtype': uuid_subtype
-        }
-
-        extra_opts['read_preference'] = kwargs.pop(
-            'read_preference',
-            self.read_preference)
-        extra_opts['tag_sets'] = kwargs.pop(
-            'tag_sets',
-            self.tag_sets)
-        extra_opts['secondary_acceptable_latency_ms'] = kwargs.pop(
-            'secondary_acceptable_latency_ms',
-            self.secondary_acceptable_latency_ms)
-
-        fields = kwargs.get('fields')
-        if fields is not None and not isinstance(fields, dict):
-            kwargs['fields'] = helpers._fields_list_to_dict(fields)
-
-        command.update(kwargs)
-
-        result = self["$cmd"].find_one(command, **extra_opts)
-
-        if check:
-            msg = "command %s failed: %%s" % repr(command).replace("%", "%%")
-            helpers._check_command_response(result, self.connection.disconnect,
-                                            msg, allowable_errors)
-
-        return result
-
-    def collection_names(self):
+    def collection_names(self, include_system_collections=True):
         """Get a list of all the collection names in this database.
+
+        :Parameters:
+          - `include_system_collections` (optional): if ``False`` list
+            will not include system collections (e.g ``system.indexes``)
         """
         results = self["system.namespaces"].find(_must_use_master=True)
         names = [r["name"] for r in results]
         names = [n[len(self.__name) + 1:] for n in names
-                 if n.startswith(self.__name + ".")]
-        names = [n for n in names if "$" not in n]
+                 if n.startswith(self.__name + ".") and "$" not in n]
+        if not include_system_collections:
+            names = [n for n in names if not n.startswith("system.")]
         return names
 
     def drop_collection(self, name_or_collection):
@@ -421,7 +476,8 @@ class Database(common.BaseObject):
 
         self.__connection._purge_index(self.__name, name)
 
-        self.command("drop", unicode(name), allowable_errors=["ns not found"])
+        self.command("drop", unicode(name), allowable_errors=["ns not found"],
+                     read_preference=ReadPreference.PRIMARY)
 
     def validate_collection(self, name_or_collection,
                             scandata=False, full=False):
@@ -459,7 +515,8 @@ class Database(common.BaseObject):
                             "%s or Collection" % (basestring.__name__,))
 
         result = self.command("validate", unicode(name),
-                              scandata=scandata, full=full)
+                              scandata=scandata, full=full,
+                              read_preference=ReadPreference.PRIMARY)
 
         valid = True
         # Pre 1.9 results
@@ -508,7 +565,8 @@ class Database(common.BaseObject):
 
         .. mongodoc:: profiling
         """
-        result = self.command("profile", -1)
+        result = self.command("profile", -1,
+                              read_preference=ReadPreference.PRIMARY)
 
         assert result["was"] >= 0 and result["was"] <= 2
         return result["was"]
@@ -548,9 +606,11 @@ class Database(common.BaseObject):
             raise TypeError("slow_ms must be an integer")
 
         if slow_ms is not None:
-            self.command("profile", level, slowms=slow_ms)
+            self.command("profile", level, slowms=slow_ms,
+                         read_preference=ReadPreference.PRIMARY)
         else:
-            self.command("profile", level)
+            self.command("profile", level,
+                         read_preference=ReadPreference.PRIMARY)
 
     def profiling_info(self):
         """Returns a list containing current profiling information.
@@ -565,7 +625,8 @@ class Database(common.BaseObject):
         Return None if the last operation was error-free. Otherwise return the
         error that occurred.
         """
-        error = self.command("getlasterror")
+        error = self.command("getlasterror",
+                             read_preference=ReadPreference.PRIMARY)
         error_msg = error.get("err", "")
         if error_msg is None:
             return None
@@ -578,7 +639,8 @@ class Database(common.BaseObject):
 
         Returns a SON object with status information.
         """
-        return self.command("getlasterror")
+        return self.command("getlasterror",
+                            read_preference=ReadPreference.PRIMARY)
 
     def previous_error(self):
         """Get the most recent error to have occurred on this database.
@@ -587,7 +649,8 @@ class Database(common.BaseObject):
         `Database.reset_error_history`. Returns None if no such errors have
         occurred.
         """
-        error = self.command("getpreverror")
+        error = self.command("getpreverror",
+                             read_preference=ReadPreference.PRIMARY)
         if error.get("err", 0) is None:
             return None
         return error
@@ -598,7 +661,8 @@ class Database(common.BaseObject):
         Calls to `Database.previous_error` will only return errors that have
         occurred since the most recent call to this method.
         """
-        self.command("reseterror")
+        self.command("reseterror",
+                     read_preference=ReadPreference.PRIMARY)
 
     def __iter__(self):
         return self
@@ -606,6 +670,80 @@ class Database(common.BaseObject):
     def next(self):
         raise TypeError("'Database' object is not iterable")
 
+    def _default_role(self, read_only):
+        if self.name == "admin":
+            if read_only:
+                return "readAnyDatabase"
+            else:
+                return "root"
+        else:
+            if read_only:
+                return "read"
+            else:
+                return "dbOwner"
+
+    def _create_or_update_user(
+            self, create, name, password, read_only, **kwargs):
+        """Use a command to create (if create=True) or modify a user.
+        """
+        opts = {}
+        if read_only or (create and "roles" not in kwargs):
+            warnings.warn("Creating a user with the read_only option "
+                          "or without roles is deprecated in MongoDB "
+                          ">= 2.6", DeprecationWarning)
+
+            opts["roles"] = [self._default_role(read_only)]
+
+        elif read_only:
+            warnings.warn("The read_only option is deprecated in MongoDB "
+                          ">= 2.6, use 'roles' instead", DeprecationWarning)
+
+        if password is not None:
+            # We always salt and hash client side.
+            if "digestPassword" in kwargs:
+                raise ConfigurationError("The digestPassword option is not "
+                                         "supported via add_user. Please use "
+                                         "db.command('createUser', ...) "
+                                         "instead for this option.")
+            opts["pwd"] = auth._password_digest(name, password)
+            opts["digestPassword"] = False
+
+        opts["writeConcern"] = self._get_wc_override()
+        opts.update(kwargs)
+
+        if create:
+            command_name = "createUser"
+        else:
+            command_name = "updateUser"
+
+        self.command(command_name, name,
+                     read_preference=ReadPreference.PRIMARY, **opts)
+
+    def _legacy_add_user(self, name, password, read_only, **kwargs):
+        """Uses v1 system to add users, i.e. saving to system.users.
+        """
+        user = self.system.users.find_one({"user": name}) or {"user": name}
+        if password is not None:
+            user["pwd"] = auth._password_digest(name, password)
+        if read_only is not None:
+            user["readOnly"] = read_only
+        user.update(kwargs)
+
+        try:
+            self.system.users.save(user, **self._get_wc_override())
+        except OperationFailure, exc:
+            # First admin user add fails gle in MongoDB >= 2.1.2
+            # See SERVER-4225 for more information.
+            if 'login' in str(exc):
+                pass
+            # First admin user add fails gle from mongos 2.0.x
+            # and 2.2.x.
+            elif (exc.details and
+                    'getlasterror' in exc.details.get('note', '')):
+                pass
+            else:
+                raise
+
     def add_user(self, name, password=None, read_only=None, **kwargs):
         """Create user `name` with password `password`.
 
@@ -634,21 +772,36 @@ class Database(common.BaseObject):
 
         .. versionadded:: 1.4
         """
-
-        user = self.system.users.find_one({"user": name}) or {"user": name}
+        if not isinstance(name, basestring):
+            raise TypeError("name must be an instance "
+                            "of %s" % (basestring.__name__,))
         if password is not None:
-            user["pwd"] = auth._password_digest(name, password)
+            if not isinstance(password, basestring):
+                raise TypeError("password must be an instance "
+                                "of %s or None" % (basestring.__name__,))
+            if len(password) == 0:
+                raise ValueError("password can't be empty")
         if read_only is not None:
-            user["readOnly"] = common.validate_boolean('read_only', read_only)
-        user.update(kwargs)
+            read_only = common.validate_boolean('read_only', read_only)
+            if 'roles' in kwargs:
+                raise ConfigurationError("Can not use "
+                                         "read_only and roles together")
 
         try:
-            self.system.users.save(user, **self._get_wc_override())
-        except OperationFailure, e:
-            # First admin user add fails gle in MongoDB >= 2.1.2
-            # See SERVER-4225 for more information.
-            if 'login' in str(e):
-                pass
+            uinfo = self.command("usersInfo", name,
+                                 read_preference=ReadPreference.PRIMARY)
+            self._create_or_update_user(
+                (not uinfo["users"]), name, password, read_only, **kwargs)
+        except OperationFailure, exc:
+            # MongoDB >= 2.5.3 requires the use of commands to manage
+            # users.
+            if exc.code in common.COMMAND_NOT_FOUND_CODES:
+                self._legacy_add_user(name, password, read_only, **kwargs)
+            # Unauthorized. MongoDB >= 2.7.1 has a narrow localhost exception,
+            # and we must add a user before sending commands.
+            elif exc.code == 13:
+                self._create_or_update_user(
+                    True, name, password, read_only, **kwargs)
             else:
                 raise
 
@@ -663,48 +816,60 @@ class Database(common.BaseObject):
 
         .. versionadded:: 1.4
         """
-        self.system.users.remove({"user": name}, **self._get_wc_override())
+
+        try:
+            self.command("dropUser", name,
+                         read_preference=ReadPreference.PRIMARY,
+                         writeConcern=self._get_wc_override())
+        except OperationFailure, exc:
+            # See comment in add_user try / except above.
+            if exc.code in common.COMMAND_NOT_FOUND_CODES:
+                self.system.users.remove({"user": name},
+                                         **self._get_wc_override())
+                return
+            raise
 
     def authenticate(self, name, password=None,
-                     source=None, mechanism='MONGODB-CR'):
+                     source=None, mechanism='MONGODB-CR', **kwargs):
         """Authenticate to use this database.
 
-        Raises :class:`TypeError` if either `name` or `password` is not
-        an instance of :class:`basestring` (:class:`str` in python 3).
         Authentication lasts for the life of the underlying client
         instance, or until :meth:`logout` is called.
 
-        The "admin" database is special. Authenticating on "admin"
-        gives access to *all* databases. Effectively, "admin" access
-        means root access to the database.
+        Raises :class:`TypeError` if (required) `name`, (optional) `password`,
+        or (optional) `source` is not an instance of :class:`basestring`
+        (:class:`str` in python 3).
 
         .. note::
-          This method authenticates the current connection, and
-          will also cause all new :class:`~socket.socket` connections
-          in the underlying client instance to be authenticated automatically.
+          - This method authenticates the current connection, and
+            will also cause all new :class:`~socket.socket` connections
+            in the underlying client instance to be authenticated automatically.
 
-         - Authenticating more than once on the same database with different
-           credentials is not supported. You must call :meth:`logout` before
-           authenticating with new credentials.
+          - Authenticating more than once on the same database with different
+            credentials is not supported. You must call :meth:`logout` before
+            authenticating with new credentials.
 
-         - When sharing a client instance between multiple threads, all
-           threads will share the authentication. If you need different
-           authentication profiles for different purposes you must use
-           distinct client instances.
+          - When sharing a client instance between multiple threads, all
+            threads will share the authentication. If you need different
+            authentication profiles for different purposes you must use
+            distinct client instances.
 
-         - To get authentication to apply immediately to all
-           existing sockets you may need to reset this client instance's
-           sockets using :meth:`~pymongo.mongo_client.MongoClient.disconnect`.
+          - To get authentication to apply immediately to all
+            existing sockets you may need to reset this client instance's
+            sockets using :meth:`~pymongo.mongo_client.MongoClient.disconnect`.
 
         :Parameters:
           - `name`: the name of the user to authenticate.
           - `password` (optional): the password of the user to authenticate.
-            Not used with GSSAPI authentication.
+            Not used with GSSAPI or MONGODB-X509 authentication.
           - `source` (optional): the database to authenticate on. If not
             specified the current database is used.
           - `mechanism` (optional): See
             :data:`~pymongo.auth.MECHANISMS` for options.
             Defaults to MONGODB-CR (MongoDB Challenge Response protocol)
+          - `gssapiServiceName` (optional): Used with the GSSAPI mechanism
+            to specify the service name portion of the service principal name.
+            Defaults to 'mongodb'.
 
         .. versionchanged:: 2.5
            Added the `source` and `mechanism` parameters. :meth:`authenticate`
@@ -725,8 +890,15 @@ class Database(common.BaseObject):
                             "of %s" % (basestring.__name__,))
         common.validate_auth_mechanism('mechanism', mechanism)
 
-        credentials = (source or self.name, unicode(name),
-                       password and unicode(password) or None, mechanism)
+        validated_options = {}
+        for option, value in kwargs.iteritems():
+            normalized, val = common.validate_auth_option(option, value)
+            validated_options[normalized] = val
+
+        credentials = auth._build_credentials_tuple(mechanism,
+                                source or self.name, unicode(name),
+                                password and unicode(password) or None,
+                                validated_options)
         self.connection._cache_credentials(self.name, credentials)
         return True
 
@@ -785,7 +957,9 @@ class Database(common.BaseObject):
         if not isinstance(code, Code):
             code = Code(code)
 
-        result = self.command("$eval", code, args=args)
+        result = self.command("$eval", code,
+                              read_preference=ReadPreference.PRIMARY,
+                              args=args)
         return result.get("retval", None)
 
     def __call__(self, *args, **kwargs):
